import os
import shutil

from .filter_nexus_for_missing import main as filter_nexus_for_missing
from .process import Process 
from subprocess import PIPE 
from .command import Command

def clean_file(in_path, out_path):

    grep_command = [
        'grep', '-oe', "SISRS_[^/]*",
        in_path,
    ]
    grep_proc = Process(grep_command, stdout=PIPE)

    uniq_command = [
        'uniq', '-c',
    ]
    uniq_proc = Process(uniq_command, stdin=grep_proc.pipe(), stdout=PIPE)

    sort_command = [
        'sort', '-k1', '-nr'
    ]
    sort_proc = Process(sort_command, stdin=uniq_proc.pipe(), stdout=PIPE)

    awk_command = [
        'awk', "{print $2}",
    ]
    awk_proc = Process(awk_command, stdin=sort_proc.pipe(), stdout=PIPE)

    # tee is simply for copying to file, since apparently awk can only deal
    # with stdout
    tee_command = [
        'tee', out_path
    ]
    tee_proc = Process(tee_command, stdin=awk_proc.pipe(), stdout=PIPE)

    tee_proc.wait()


    # TODO: fix python implementation
    # For some reason the python code below has some subtle differences in
    # the output files at the binary level. The text is the same but the hex
    # dumps are different

    #with open(in_path, 'r') as f_in:
    #    
    #    lines = [ line[:line.find('/')] for line in f_in ]

    #counts = {}
    #for name in lines:
    #    if name not in counts:
    #        counts[name] = 0
    #    counts[name] += 1

    #print(counts)

    #sorted_counts = sorted(counts.items(), key=lambda x: x[1], reverse=True)

    #with open(out_path, 'w') as f_out:

    #    for count in sorted_counts:
    #        f_out.write("{}\n".format(count[0]))

def run_comparison(args, str_missing, locs_filename, locs_clean_filename,
        dirname, filename):

    dir_ = os.path.join(
            args['out_dir'],
            dirname)

    alignment_file_path = os.path.join(dir_, filename)

    # TODO: should we really be checking if the directory exists first?
    if not os.path.exists(dir_):
        os.makedirs(dir_)
        src_path = os.path.join(args['out_dir'], filename)
        shutil.move(src_path, alignment_file_path)

    filter_nexus_for_missing(alignment_file_path, str_missing)

    locs_file_path = os.path.join(dir_, locs_filename)
    locs_clean_file_path = os.path.join(dir_, locs_clean_filename)
    
    clean_file(locs_file_path, locs_clean_file_path)


class ChangeMissingCommand(Command):

    def run(self):

        args = self._args
        missing = args['missing']

        alignment_file_path = os.path.join(args['out_dir'], 'alignment.nex')

        if missing is None:
            dir_lists = args['dir_lists']
            all_dirs = dir_lists.get_all_dirs()
            str_missing = str(len(all_dirs) - 2)
        else:
            str_missing = str(missing)

        filter_nexus_for_missing(alignment_file_path, str_missing)

        locs_filename = 'locs_m{}.txt'.format(str_missing)
        locs_file_path = os.path.join(args['out_dir'], locs_filename)

        locs_clean_filename = 'locs_m{}_Clean.txt'.format(str_missing)
        locs_clean_file_path = os.path.join(
            args['out_dir'], locs_clean_filename)

        clean_file(locs_file_path, locs_clean_file_path)

        run_comparison(
                args,
                str_missing,
                locs_filename,
                locs_clean_filename,
                'alignmentDataWithoutSingletons',
                'alignment_pi.nex')

        run_comparison(
                args,
                str_missing,
                locs_filename,
                locs_clean_filename,
                'alignmentDataWithOnlyBiallelic',
                'alignment_bi.nex')




import os
import sys
from .process import Process 
from multiprocessing import Pool
from .specific_genome import main as specific_genome
from .get_pruned_dict import main as get_pruned_dict
from .aligners import create_aligner
from .command import Command

def run_mpileup(args):

    dir_ = args[0]
    contig_file_path = args[1]
    
    taxon_name = os.path.basename(dir_)

    path = os.path.join(dir_, taxon_name)

    print(path)

    command = [
        'samtools', 'mpileup',
        '-f', contig_file_path,
        '-o', path + '.pileups',
        path + '.bam', 
    ]
    Process(command).wait()

def run_specific_genome(args):

    dir_ = args[0]
    contig_file_path = args[1]

    specific_genome(dir_, contig_file_path)

def run_faidx(dir_):

    path = os.path.join(dir_, 'contigs.fa')

    command = [
        'samtools', 'faidx', path
    ]
    Process(command).wait()

def run_bowtie2(args):

    aligner = args[0]
    dir_ = args[1]

    contigs_path = os.path.join(dir_, 'contigs.fa')
    contigs_dir = os.path.join(dir_, 'contigs')

    aligner.index(contigs_path, contigs_dir)

def run_index(dir_):

    taxon_name = os.path.basename(dir_)
    bam_path = os.path.join(dir_, taxon_name + '.bam')

    command = [
        'samtools', 'index',
        bam_path,
    ]
    Process(command).wait()

def run_get_pruned_dict(args):
    get_pruned_dict(*args)


class IdentifyFixedSitesCommand(Command):

    def run(self):

        args = self._args

        contig_dir = args['contig_dir']
        num_processors = args['num_processors']
        dir_lists = args['dir_lists']
        min_read = args['min_read']
        threshold = args['threshold']

        aligner = create_aligner(num_processors=num_processors)

        all_dirs = dir_lists.get_all_dirs()

        contig_file_path = os.path.join(contig_dir, 'contigs.fa')

        Process(['samtools', 'faidx', contig_file_path]).wait()

        pool = Pool(num_processors)
        args = [ (dir_, contig_file_path) for dir_ in all_dirs ]

        pool.map(run_mpileup, args)

        try:
            pool.map(run_specific_genome, args)
        except Exception as e:
            print("specific_genome.py failed")
            print(e)
            sys.exit(1)

        pool.map(run_faidx, all_dirs)
        pool.map(run_bowtie2, [(aligner, dir_) for dir_ in all_dirs])

        for dir_ in all_dirs:

            taxon_name = os.path.basename(dir_)
            os.remove(os.path.join(dir_, taxon_name + '.bam'))
            os.remove(os.path.join(dir_, taxon_name + '.bam.bai'))
            os.remove(os.path.join(dir_, taxon_name + '.pileups'))

            contig_prefix = os.path.join(dir_, 'contigs')
            aligner.align(dir_, contig_prefix)

        args = []
        for dir_ in all_dirs:
            contig_file_path = os.path.join(dir_, 'contigs.fa')
            args.append([dir_, contig_file_path])

        pool.map(run_index, all_dirs)

        pool.map(run_mpileup, args)
 
        # put base for each site in a dictionary (allows no variation when
        # calling sites)
        args = [ (dir_, min_read, threshold) for dir_ in all_dirs ]
        try:
            pool.map(run_get_pruned_dict, args)
        except Exception as e:
            print("get_pruned_dict.py failed")
            print(e)
            sys.exit(1)
        print("==== Done Identifying Fixed Sites Without Error ====")



import os
import sys
from multiprocessing import Pool

from .command import Command
from .sub_sample_for_velvet_unshuff import main as sub_sample_for_velvet_unshuff

def run_subsample_on_dir(args):
    dir_ = args[0]
    left_reads = args[1]
    sub_sample_for_velvet_unshuff(dir_, left_reads, '_R1', '_R2')

class SubsampleCommand(Command):

    def run(self):

        args = self._args

        genome_size = args['genome_size']
        if genome_size is None:
            print("Error: genome size required for subsampling")
            sys.exit(1)

        print("Subsampling data")
        out_dir = args['out_dir']
        dir_lists = args['dir_lists']
        all_dirs = dir_lists.get_all_dirs()
        num_processors = args['num_processors']

        subsample_dir = os.path.join(out_dir, 'subsamples')
        if os.path.exists(subsample_dir):
            os.remove(subsample_dir)
        os.makedirs(subsample_dir)

        numer = 10 * genome_size
        # assumes 100bp reads - add option in future
        denom = 100 * 2 * len(all_dirs)
        left_reads = numer / denom

        pool_args = [ (dir_, left_reads) for dir_ in all_dirs ]
        pool = Pool(num_processors)

        try:
            pool.map(run_subsample_on_dir, pool_args)
        except Exception as e:
            print("sub_sample_for_velvet_unshuff.py failed")
            print(e)
            sys.exit(1)




import os
from subprocess import Popen
from glob import glob
from subprocess import Popen, PIPE 


class Process(object):

    def __init__(self, command, stdin=None, stdout=None):
        try:
            self.proc = Popen(command, stdin=stdin, stdout=stdout)
        except Exception as e:
            print("Error calling: {}".format(command[0]))
            raise e 

    def wait(self):
        return self.proc.wait()

    def pipe(self):
        return self.proc.stdout


class MpileupProcess(object):

    def __init__(self, dir_, contig_file_path, num_processors):

        taxon_name = os.path.basename(dir_)
        bam_path = os.path.join(dir_, '{}.bam'.format(taxon_name))
        pileups_path  = os.path.join(dir_, '{}.pileups'.format(taxon_name))

        command = [
            'samtools',
            '-f', contig_file_path,
            bam_path,
            '-o', pileups_path
        ]
        Process(command).wait()

    #parallel --jobs "${PROCESSORS}" 'samtools mpileup -f' "${OUTFOLDER}"/"${CONTIGS}"/contigs.fa '"$( echo {}/$(basename {} ) )".bam' '> "$( echo {}/$(basename {} ) )".pileups' ::: "${FOLDERLISTA[@]}"



#!/usr/bin/env python3

"""Take multiple mpileup files and determine the base at each position of the reference genome

    arguments:
        path: folder containing mpileup files ending in pileups
        minread: number of reads at a position required to call a base
        thresh: proportion of reads that must be one base for calling to occur

    output:
        path/pruned_dict.pkl : contains pickled dictionary of position:base

"""

import sys
# NOTE: python 3 uses cPickle behind the scenes by default
import pickle as cPickle
from collections import Counter
import glob
import string
import re
import os
from .specific_genome import getCleanList

#get combined pileup info
def getallbases(path,minread,thresh):
    assert len(glob.glob1(path,"*.pileups"))==1,'More than one pileup file in'+path
    allbases=dict()
    with open (path+'/'+os.path.basename(path)+'.pileups',"r") as filein:
        for line in filein:
            splitline=line.split()
            if len(splitline)>4:
                node,pos,ref,num,bases,qual=line.split()
                loc=node+'/'+pos
                cleanBases=getCleanList(ref,bases)  #Get clean bases where * replaced with -
                assert len(cleanBases) == int(num), 'bases are being counted incorrectly: '+ str(bases) + ' should have '+str(num)+' bases, but it is being converted to '+"".join(cleanBases)
                finalBase=getFinalBase_Pruned(cleanBases,minread,thresh)
                if finalBase != 'N':    #Do not pass Ns to pruned_dictionary, but do pass - as deletion
                    allbases[loc]=finalBase
    return allbases

def getFinalBase_Pruned(cleanBases,minread,thresh):
    singleBase=(Counter(cleanBases).most_common()[0][0])
    if singleBase == '*':
        singleBase = '-'
    counts=int((Counter(cleanBases).most_common()[0][1]))

    if counts >= minread and counts/float(len(cleanBases)) >= thresh:
        finalBase=singleBase
    else:
        finalBase='N'

    return finalBase
###############################################
def main(path, minread, thresh):

    allbases=getallbases(path,minread,thresh)      #dictionary of combined pileups - locus/pos:bases(as list)
    if len(allbases)==0:
        print('No data for '+path)
        sys.exit(1)      #dictionary of combined pileups - locus/pos:bases(as list)

    output = open(path+'/pruned_dict.pkl', 'wb')
    # Use protocol version 2 in order to match the test data which was
    # generated using the bash version of SISRS running on Python 2. This
    # can probably be changed to cPickle.HIGHEST_PROTOCOL but the test data
    # will need to be updated. This was the best way I could think of to make
    # minimal changes to bash SISRS in order to create a base set of test
    # data.
    cPickle.dump(allbases, output, 2)
    #cPickle.dump(allbases, output, cPickle.HIGHEST_PROTOCOL)
    output.close()


if __name__ == '__main__':
    path = sys.argv[1]
    minread = int(sys.argv[2])
    thresh = float(sys.argv[3])
    main(path, minread, thresh)



from abc import ABCMeta, abstractmethod

class Aligner(metaclass=ABCMeta):

    def __init__(self, num_processors=1):
        self._num_processors = num_processors

    @abstractmethod
    def index():
        """Create index"""

    @abstractmethod
    def align():
        """Perform alignment"""



#!/usr/bin/env python3

"""Take a nex or phy or fa file, outputs phylip-relaxed with X missing data

    arguments:
    data  -- extension should be nex/phy/fa
    missing -- the number of species in the alignment allowed to have missing data

    output:
    phylip formatted file ending with _mX.phylip-relaxed where X is the number missing
    """

from __future__ import division
import sys
from os import path
import linecache
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio.Alphabet import generic_dna, IUPAC, Gapped
from Bio.Align import MultipleSeqAlignment, AlignInfo
from Bio import AlignIO, SeqIO

def main(alignment_filename, missing_str):

    ######################
    bases = ['A', 'C', 'G', 'T', 'a', 'c', 'g', 't','-']
    formats = {'nex':'nexus', 'phy':'phylip-relaxed', 'fa':'fasta'}
    fformat = formats[alignment_filename.split('.')[-1]]
    missing = int(missing_str)
    data = SeqIO.to_dict(SeqIO.parse(alignment_filename, fformat))
    locline = linecache.getline(alignment_filename, 8)
    locs = locline.split()
    species = list(data.keys())
    minsp = len(species)-missing
    newlocs = list()

    for k in data:
        data[k] = list(data[k].seq)

    newdata = {sp: list() for sp in species}
    for i in range(len(data[species[0]])):
        site = [data[sp][i] for sp in species if data[sp][i] in bases]
        if len(set(site)) > 1 and len(site) >= minsp:
            for sp in species:
                newdata[sp].append(data[sp][i])
            newlocs.append(locs[i+1])

    datalist = []
    for k,v in sorted(newdata.items()):
        seq = SeqRecord(Seq(''.join(v)), id=k)
        datalist.append(seq)

    SeqIO.write(datalist, path.dirname(alignment_filename)+'/'+path.basename(alignment_filename).split('.')[0]+'_m'+missing_str+'.phylip-relaxed', "phylip-relaxed")
    locfile = open(path.dirname(alignment_filename)+'/locs_m'+missing_str+'.txt', 'w')
    locfile.write("\n".join(newlocs))
    locfile.close()
    origLength = len(data[species[0]])
    newLength = len(newlocs)
    print('With '+str(missing)+' taxa allowed to be missing, '+str(origLength)+' sites from '+path.basename(alignment_filename)+' ('+str(len(species)-2)+' allowed missing) are reduced to '+str(len(newlocs))+' sites ('+str(origLength-newLength)+' sites or '+str('%.2f' % (((origLength-newLength)/origLength)*100))+'% lost)')


if __name__ == '__main__':
    alignment_filename = sys.argv[1]
    missing_str = sys.argv[2]

    main(alignment_filename, missing_str)



from multiprocessing import Pool
import os
import sys
import argparse
from glob import glob
from pprint import pprint
from .subsample import SubsampleCommand
from .align_contigs import AlignContigsCommand 
from .identify_fixed_sites import IdentifyFixedSitesCommand
from .output_alignment import OutputAlignmentCommand
from .change_missing import ChangeMissingCommand
from .process import Process


class SISRSPipeline(object):
    def __init__(self):
         
        self._pipeline = [
            [ 'subSample', self._subsample ],
            [ 'alignContigs', self._align_contigs ],
            [ 'identifyFixedSites', self._identify_fixed_sites ],
            [ 'outputAlignment', self._output_alignment],
            [ 'changeMissing', self._change_missing],
        ]

        self._pipeline_lookup = {
            stage[0]: {
                'index': i,
                'func': stage[1]
            } for i, stage in enumerate(self._pipeline)
        }

    def get_command_list(self):
        commands = []
        for stage in self._pipeline:
            commands.append(stage[0])
        return commands

    def run_command(self, command_name, args):
        self._pipeline_lookup[command_name]['func'](args)

    def run_commands_from(self, first_command_name, args):
        
        first_command_index = \
            self._pipeline_lookup[first_command_name]['index']

        for stage in self._pipeline[first_command_index:]:
            command_name = stage[0]
            self.run_command(command_name, args)

    @staticmethod
    def _subsample(args):
        command = SubsampleCommand(args)
        command.run()

    @staticmethod
    def _align_contigs(args):
        command = AlignContigsCommand(args)
        command.run()

    @staticmethod
    def _identify_fixed_sites(args):
        command = IdentifyFixedSitesCommand(args)
        command.run()
        
    @staticmethod
    def _output_alignment(args):
        command = OutputAlignmentCommand(args)
        command.run()

    @staticmethod
    def _change_missing(args):
        command = ChangeMissingCommand(args)
        command.run()


class DirectoryLists(object):

    def __init__(self, base_dir):

        self._all_fastq = []
        self._paired = []
        self._unpaired = []
        
        for root, dirs, files in os.walk(base_dir):
            for filename in files:

                file_path = os.path.join(root, filename)

                if filename.endswith('.fastq'):
                    self._all_fastq.append(file_path)

                if self._is_paired_read_filename(filename):
                    self._paired.append(file_path)
                else:
                    self._unpaired.append(file_path)

        self._all_dirs = sorted(list(set([ os.path.dirname(x) for x in self._all_fastq ])))

    def get_all_dirs(self):
        return self._all_dirs

    def get_paired(self):
        return self._paired

    def get_unpaired(self):
        return self._unpaired

    def _is_paired_read_filename(self, filename):
        return (filename.endswith('_R1.fastq') or
                filename.endswith('_R2.fastq') or
                # TODO: is it true that subsampled files are always paired?
                'subsampled' in filename)


def setup_output_directory(data_directory, output_directory, overwrite):

    if output_directory is None:
        output_directory = data_directory
        print("Note: SISRS writing into data folder")
    elif not os.path.exists(output_directory):
        print(data_directory)
        recursive_symlinks(data_directory, output_directory)
    else:
        if overwrite:
            print("{} already exists. Overwriting...".format(
                output_directory))
            recursive_symlinks(data_directory, output_directory)
        else:
            print("{} already exists and overwrite flag not set. Aborting...".format(
                output_directory))
            sys.exit(1)

    return os.path.abspath(output_directory)

# based off https://stackoverflow.com/a/15824216/943814
def recursive_symlinks(src, dest):
    if os.path.isdir(src):
        if not os.path.isdir(dest):
            os.makedirs(dest)
        files = glob(os.path.join(src, '*'))
        for f in files:
            f = os.path.basename(f)
            next_src = os.path.join(src, f)
            next_dest = os.path.join(dest, f)
            recursive_symlinks(
                next_src, next_dest)
    else:
        try:
            os.symlink(src, dest)
        except OSError as e:
            os.remove(dest)
            os.symlink(src, dest)


def main():

    pipeline = SISRSPipeline()

    parser = argparse.ArgumentParser()

    parser.add_argument(
            'command',
            choices=pipeline.get_command_list(),
            help="Command to run")
    parser.add_argument('-c', '--continuous', default=1)
    parser.add_argument('-f', '--data-directory')
    parser.add_argument('-z', '--output-directory')
    parser.add_argument('--overwrite')
    parser.add_argument('-a', '--assembler', type=str, default='velvet')
    parser.add_argument('-p', '--num_processors', type=int, default=1)
    parser.add_argument('-m', '--missing', type=int, help="Num missing")
    parser.add_argument('-g', '--genome-size', type=int, help="Genome size",
            default=None)

    # identifyFixedSites specific
    parser.add_argument(
            '-n', '--min-read', required=False, type=int, default=3,
            help='Min read')
    parser.add_argument(
            '-t', '--threshold', required=False, type=int, default=1,
              help="Threshold for calling the site")

    args = parser.parse_args()

    data_directory = args.data_directory
    output_directory = args.output_directory
    overwrite = args.overwrite
    assembler = args.assembler
    num_processors = args.num_processors
    continuous = args.continuous

    if data_directory is None:
        data_directory = os.getcwd()
    data_directory = os.path.abspath(data_directory)

    output_directory = setup_output_directory(
        data_directory, output_directory, overwrite)

    dir_lists = DirectoryLists(output_directory)

    contig_dir = ''
    if assembler == 'premade':
        contig_dir = 'premadeoutput'

    args_dict = {}
    args_dict['data_dir'] = data_directory
    args_dict['out_dir'] = output_directory
    args_dict['assembler'] = assembler
    args_dict['dir_lists'] = dir_lists
    args_dict['contig_dir'] = os.path.join(output_directory, contig_dir)
    args_dict['num_processors'] = num_processors
    args_dict['continuous'] = continuous
    args_dict['min_read'] = args.min_read 
    args_dict['threshold'] = args.threshold
    args_dict['missing'] = args.missing
    args_dict['genome_size'] = args.genome_size

    command_name = args.command

    if continuous == 1:
        pipeline.run_commands_from(command_name, args_dict)
    else:
        pipeline.run_command(command_name, args_dict)




#!/usr/bin/env python3

"""
    output alignment of sites useful for phylogenetics (nexus format)
    can specify max number of species missing for each site

    arguments:
        num_missing -- the number of species allowed to be missing at a site so that site will show in output alignment
        reference genome -- if 'X' then output will only show results with reference to composite-reference; otherwise will also show position given alignment of composite-reference to an assembled genome
        mainfolder -- folder that includes folders containing data
        assembler -- assembler used to get composite reference

    output:
        alignment.nex : nexus formatted alignment including position in composite reference genome (and absolute ref if available); each site as up to num_missing missing data
        alignment_bi.nex : above but only biallelic sites
        alignment_pi.nex : above but only phylogenetically informative sites (no singletons)

"""

import os
import re
import glob
# NOTE: Python 3 imports cPickle behind the scenes by default
import pickle as cPickle
from collections import Counter,defaultdict
import sys
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio import SeqIO

#########################
class Scaffold:
    def __init__(self,chr,start,end,length=None,flag=None,cigar=None):
        self.chr = chr
        self.start = start
        self.end = end
        self.length = length
        self.flag = flag
        self.cigar = cigar

class Loc:
    def __init__(self,scaff_loc,flag):
        self.scaff_loc = scaff_loc
        self.flag = flag

class Alignment:
    def __init__(self,locations=[],species_data=dict(),ref=[],ref_loc=[],flag=[],single=[]):
        self.locations = locations
        self.species_data = species_data
        self.ref = ref
        self.ref_loc = ref_loc
        self.flag = flag
        self.single = single

    def numsnps(self):
        print(str(len(self.locations))+' total variable sites (alignment.nex)')
        for i in range(len(self.locations)):
            bases = [self.species_data[sp][i] for sp in self.species_data if self.species_data[sp][i] in ['A','C','G','T','-']]     #bases for that site
            c = Counter(bases).most_common(5)
            if c[1][1]==1:
                self.single.append(1)
            else:
                self.single.append(0)
            self.flag.append(len(c))

        print(str(self.single.count(1))+' variable sites are singletons')

        return self.flag.count(2)       # number of biallelic sites

def makerefdict(reffasta):
    ''' takes fasta file, returns dict of contig_name:[sequence]    '''

    filein=open(reffasta, "rU") #read fasta file
    refdict = SeqIO.to_dict(SeqIO.parse(filein, "fasta"))
    filein.close()

    for k in refdict:
        refdict[k]=list(str(refdict[k].seq))

    return refdict      #name:sequence_as_list

def read_pkls(path):
    '''     read python dicts back from the files
            get species:{loc:base}
                list of locs
                list of species folders     '''

    filelist = glob.glob(path+'/*/pruned_dict.pkl')
    assert len(filelist) > 0, 'No species had data from the pileup'

    allbases=dict()
    alllocs,pathlist=[],[]

    for fi in filelist:
        d = os.path.dirname(fi)     #sp relative path
        pathlist.append(d)
        species = os.path.basename(d)
        print('Reading data: '+ species)     #print sp name

        pkl_file = open(fi, 'rb')
        sp_bases = cPickle.load(pkl_file)       #sp_bases is loc:base for species
        pkl_file.close()

        for l in sp_bases:
            alllocs.append(l)
        allbases[species]=sp_bases

    alllocs=list(set(alllocs))
    pathlist.sort()
    alllocs.sort()

    return allbases,alllocs,pathlist

def get_phy_sites(pathlist,allbases,alllocs,num_missing):
    ''' gets the alignment i.e. the list of bases for each species, where all sites are variable among species  '''

    splist = [os.path.basename(path) for path in pathlist]
    alignment = Alignment()
    alignment.species_data = {species: [] for species in splist}

    for loc in alllocs: #go through each location
        snp = [allbases[species][loc] for species in splist if loc in allbases[species]]      #list of the base for each species at a given loc
        snp = [b for b in snp if b in ['A','G','C','T','-']]        #filter for real bases and deletions

        if (len(pathlist)-len(snp)) <= num_missing and len(set(snp)) > 1:     #not too many missing and there is variation
            alignment.locations.append(loc)
            for species in splist: #add base to the list for that species (or add 'N')
                if loc in allbases[species]:
                    alignment.species_data[species].append(allbases[species][loc])
                else:
                    alignment.species_data[species].append('N')

    return alignment

def sep_cigar(cigar):
    sepcigar = re.findall('[A-Z]+|[0-9]+',cigar)
    return sepcigar[::2], sepcigar[1::2]       #nums, code

def adjust_mapping(readmap,cigar,flag):
    nums,code = sep_cigar(cigar)

    i=1
    newreadmap = readmap

    if flag == 1:
        code.reverse()
        nums.reverse()

    for j,num in enumerate(nums):
        if code[j] == 'S' or code[j] == 'I':
            newreadmap = newreadmap - num
        elif code[j] == 'D':
            newreadmap = newreadmap + num

        i = i+num
        if i>readmap:
            break

    return newreadmap

def add_ref_info(alignment,ref,nodedict):
    if ref == 0:
        alignment.ref_loc = ['X'] * len(alignment.locations)        #no reference, just use X for all ref locations
    else:
        for loc in alignment.locations:
            node,site=loc.split('/')
            pos=int(site)
            if nodedict[node] is not 'X': #contig maps to reference
                chro=nodedict[node].chr
                contig_map=nodedict[node].start
                flag = nodedict[node].flag
                cigar = nodedict[node].cigar
                if flag == '1':
                    pos=nodedict[node].length-pos+1
                pos = adjust_mapping(pos,cigar,0)
                pos=contig_map+pos-1   #1 based position
                alignment.ref_loc.append(chro+'_'+str(pos)) #add chr_pos
                if (pos-1) < len(ref[chro]):
                    refbase=ref[chro][(pos-1)] #chr in dictionary, site within seq-as-list   0 based position
                    if flag == '1':
                        if refbase in basecomplement:
                            refbase = basecomplement[refbase]
                        else:
                            refbase = 'N'
                    alignment.ref.append(refbase.upper()) #add ref base
            else:
                alignment.ref_loc.append('X')
                alignment.ref.append('N')

    return alignment

def get_ref_info(alignment,reference,mainfolder,assembler):
    ''' #identify the chromosome and starting position for each contig '''

    nodedict = dict((loc.split('/')[0],'X') for loc in set(alignment.locations))    #make each contig name a key with an empty value (if it has a snp)
    if reference is 'X':
        ref=0
    else:
        ref=makerefdict(reference)        #make dict for reference chromosomes
        samfile=open(mainfolder+'/'+assembler+'output/align_contigs.sam','r') #sam file contains alignment of contigs to reference ref genome - has chr and pos
        for line in samfile:
            if not line.startswith('@'):    #skip header
                splitline=line.split()
                if splitline[0] in nodedict:        #only interested if contig has a snp
                    flag = bin(int(splitline[1]))
                    flagl = flag.split('b')
                    if len(flagl[1])>=3:
                        if flagl[1][-3] == '1':
                            continue                #if contig is unmapped, continue
                    if len(flagl[1])>=5:
                        flag = flagl[1][-5]     #check for reverse mapping
                    else:
                        flag = '0'

                    chro=splitline[2]
                    start=int(splitline[3])
                    cigar=(splitline[5])
                    end = start + (adjust_mapping(len(splitline[9]),cigar,0)) - 1
                    nodedict[splitline[0]]=Scaffold(chro,start,end,len(splitline[9]),flag,cigar)
    return nodedict,ref

def write_alignment(fi,alignment,numbi):
    spp = sorted(alignment.species_data.keys())

    if len(alignment.ref) == 0:
        ntax = str(len(alignment.species_data))
    else:
        ntax = str(len(alignment.species_data)+1)

    ALIGNMENT=open(fi,'w')
    ALIGNMENTBI=open(fi.replace('.nex','_bi.nex'),'w')
    ALIGNMENTPI=open(fi.replace('.nex','_pi.nex'),'w')

    ALIGNMENT.write('#NEXUS\n\nBEGIN DATA;\nDIMENSIONS NTAX='+ntax+' NCHAR='+str(len(alignment.locations))+';\nFORMAT MISSING=? GAP=- DATATYPE=DNA;\nMATRIX\n')
    ALIGNMENT.write('[ '+ " ".join(alignment.ref_loc)+' ]'+"\n")
    ALIGNMENT.write('[ '+ " ".join(alignment.locations)+' ]'+"\n")
    for species in spp: #write sequences for each species
        ALIGNMENT.write(species+"\t"+"".join(alignment.species_data[species])+"\n")


    bi_ref_loc = [alignment.ref_loc[i] for i in range(len(alignment.locations)) if alignment.flag[i] == 2 and alignment.single[i] == 0]
    bi_loc = [alignment.locations[i] for i in range(len(alignment.locations)) if alignment.flag[i] == 2 and alignment.single[i] == 0]
    bi_sp_data={}
    for species in spp:
        bi_sp_data[species] = [alignment.species_data[species][i] for i in range(len(alignment.locations)) if alignment.flag[i] == 2 and alignment.single[i] == 0]
    if len(alignment.ref) > 0:
        bi_sp_data['reference'] = [alignment.ref[i] for i in range(len(alignment.locations)) if alignment.flag[i] == 2 and alignment.single[i] == 0]

    ALIGNMENTBI.write('#NEXUS\n\nBEGIN DATA;\nDIMENSIONS NTAX='+ntax+' NCHAR='+str(len(bi_loc))+';\nFORMAT MISSING=? GAP=- DATATYPE=DNA;\nMATRIX\n')
    ALIGNMENTBI.write('[ '+ " ".join(bi_ref_loc)+' ]'+"\n")
    ALIGNMENTBI.write('[ '+ " ".join(bi_loc)+' ]'+"\n")
    for species in spp: #write sequences for each species
        ALIGNMENTBI.write(species+"\t"+("".join(bi_sp_data[species]))+"\n")
    print(str(len(bi_loc))+' total biallelic sites excluding singletons (alignment_bi.nex)')

    pi_ref_loc = [alignment.ref_loc[i] for i in range(len(alignment.locations)) if alignment.single[i] == 0]
    pi_loc = [alignment.locations[i] for i in range(len(alignment.locations)) if alignment.single[i] == 0]
    pi_sp_data={}
    for species in spp:
        pi_sp_data[species] = [alignment.species_data[species][i] for i in range(len(alignment.locations)) if alignment.single[i] == 0]
    if len(alignment.ref) > 0:
        pi_sp_data['reference'] = [alignment.ref[i] for i in range(len(alignment.locations)) if alignment.single[i] == 0]

    ALIGNMENTPI.write('#NEXUS\n\nBEGIN DATA;\nDIMENSIONS NTAX='+ntax+' NCHAR='+str(len(pi_loc))+';\nFORMAT MISSING=? GAP=- DATATYPE=DNA;\nMATRIX\n')
    ALIGNMENTPI.write('[ '+ " ".join(pi_ref_loc)+' ]'+"\n")
    ALIGNMENTPI.write('[ '+ " ".join(pi_loc)+' ]'+"\n")
    for species in spp: #write sequences for each species
        ALIGNMENTPI.write(species+"\t"+("".join(pi_sp_data[species]))+"\n")
    print(str(len(pi_loc))+' total variable sites excluding singletons (alignment_pi.nex)')

    if len(alignment.ref) > 0:
        ALIGNMENT.write('reference'+"\t"+("".join(alignment.ref))+"\n")

    ALIGNMENT.write(';\nend;')
    ALIGNMENTBI.write(';\nend;')
    ALIGNMENTPI.write(';\nend;')

    ALIGNMENT.close()
    ALIGNMENTBI.close()
    ALIGNMENTPI.close()

#########################

def main(num_missing, reference, mainfolder, assembler):
    basecomplement = {'a':'t', 'c':'g', 't':'a', 'g':'c', 'A':'t', 'C':'g', 'T':'a', 'G':'c','-':'-'}

    allbases,alllocs,pathlist = read_pkls(mainfolder)      #dict of species:(loc:base), sorted list of unique position names
    num_species = len(pathlist)

    alignment=get_phy_sites(pathlist,allbases,alllocs,num_missing)       #return Alignment object of informative sites
    numbi = alignment.numsnps()     #prints numbers of snps, biallelic snps, and singletons

    nodedict,ref = get_ref_info(alignment,reference,mainfolder,assembler)

    alignment = add_ref_info(alignment,ref,nodedict)
    alignment = write_alignment(mainfolder+'/alignment.nex',alignment,numbi)


if __name__ == '__main__':
    num_missing = int(sys.argv[1])
    reference = sys.argv[2]
    mainfolder = sys.argv[3]
    assembler = sys.argv[4]
    main(num_missing, reference, mainfolder, assembler)



import os
import shutil
from multiprocessing import Pool
from subprocess import check_call
from .process import Process 
from .aligners import create_aligner
from .command import Command


def sam_index_directory(dir_):
    basename = os.path.basename(dir_)
    filename = basename + '.bam'
    file_path = os.path.join(dir_, filename)

    index_command = [
        'samtools', 'index', file_path
    ]
    index_proc = Process(index_command)
    index_proc.wait()


class AlignContigsCommand(Command):

    def run(self):

        print("==== Renaming Scaffolds for SISRS ====")
        contig_dir = self._args['contig_dir']
        dir_lists = self._args['dir_lists']
        num_processors = self._args['num_processors']

        aligner = create_aligner(num_processors=num_processors)

        contig_file_path = os.path.join(contig_dir, 'contigs.fa')

        # backup contigs.fa
        backup_contig_file_path = os.path.join(
            contig_dir, 'contigs_OriginalNames.fa')
        shutil.move(contig_file_path, backup_contig_file_path)

        # rename scaffolds
        rename_command = [
            'rename.sh',
            'in={}'.format(backup_contig_file_path),
            'out={}'.format(contig_file_path),
            'prefix=SISRS',
            'addprefix=t'
        ]
        check_call(rename_command)
        print("==== Scaffolds Renamed ====")

        all_dirs = dir_lists.get_all_dirs()
        contig_prefix = os.path.join(contig_dir, 'contigs')

        aligner.index(contig_file_path, contig_prefix)

        for dir_ in all_dirs:

            aligner.align(dir_, contig_prefix)

        print("==== Done Aligning ====")
        pool = Pool(num_processors)
        pool.map(sam_index_directory, all_dirs)
        print("==== Done Indexing Bam Files ====")





#!/usr/bin/env python3
#This script assumes a single pileup file per taxon
#Add an assert statement to test this
import os
import sys
from collections import Counter
from Bio import SeqIO
import glob

#get combined pileup info
def getallbases(path):
    assert len(glob.glob1(path,"*.pileups"))==1,'More than one pileup file in'+path
    allbases=dict()
    with open (path+'/'+os.path.basename(path)+'.pileups',"r") as filein:
        for line in filein:
            splitline=line.split()
            if len(splitline)>4:
                node,pos,ref,num,bases,qual=line.split()
                loc=node+'/'+pos
                cleanBases=getCleanList(ref,bases)
                assert len(cleanBases) == int(num), 'bases are being counted incorrectly: '+ str(bases) + ' should have '+str(num)+' bases, but it is being converted to '+"".join(cleanBases)
                finalBase=getFinalBase_Specific(cleanBases)
                allbases[loc]=finalBase
    return allbases

def getCleanList(ref,bases):
    bases=bases.replace('.',ref) #insert ref base
    bases=bases.replace(',',ref)
    bases=bases.upper() #everything in uppercase
    bases=list(bases)

    okbases=['A','C','G','T','*']
    indels=['+','-']

    new_base_list=[]
    ibase_list = iter(bases)
    for b in ibase_list:
        if b in okbases:
            new_base_list.append(b)         #Get base
        elif b in indels:                   #skip indels
            i = int(next(ibase_list))
            j = str(next(ibase_list))
            if str.isdigit(j):
                skip=int(str(i)+j)
                while skip>0:
                    z=next(ibase_list)
                    skip=skip-1
            else:
                while i>1:
                    z=next(ibase_list)
                    i = i-1
        elif b=='^':                        #skip read qual noted at end of read
            z=next(ibase_list)

    return new_base_list

def getFinalBase_Specific(cleanBases):
    most_common = sorted(Counter(cleanBases).most_common())
    finalBase=(most_common[0][0])
    if finalBase == '*':
        finalBase = 'N'
    return finalBase

###############################################
#if __name__ == "__main__":
#    path=sys.argv[1]
#    contig_file = sys.argv[2]

def main(path, contig_file):

    allbases=getallbases(path)      #dictionary of combined pileups - locus/pos:bases(as list)
    if len(allbases)==0:
        print('No data for '+path)
        sys.exit(1)

    # Read contig fasta file into dictionary with sequence ID as the key
    contig_handle = open(contig_file, "r")
    fasta_seq = SeqIO.parse(contig_handle, 'fasta')
    fasta_dict = {read.id:list(str(read.seq)) for read in fasta_seq}
    contig_handle.close()

    for locus_pos,base in sorted(allbases.items()):
        locus,pos = locus_pos.split('/')
        fasta_dict[locus][int(pos)-1] = base

    #output = open(path+'/contigs.fa', 'wb')
    output = open(path+'/contigs.fa', 'w')
    for l,seq in sorted(fasta_dict.items()):
        output.write('>'+str(l)+"\n"+"".join(seq)+"\n")
    output.close()


if __name__ == '__main__':
    path = sys.argv[1]
    contig_file = sys.argv[2]
    main(path, contig_file)



from .command import Command
from .get_alignment import main as get_alignment


class OutputAlignmentCommand(Command):

    def run(self):
        args = self._args

        data_dir = args['data_dir']
        out_dir = args['out_dir']
        assembler = args['assembler']
        dir_lists = args['dir_lists']

        all_dirs = dir_lists.get_all_dirs()
        num_missing = len(all_dirs) - 2
        get_alignment(num_missing, 'X', out_dir, assembler)





#!/usr/bin/env python2

""" Subsample fastq data using reservoir sampling
    assumes reads are 100 bp
    samples read pairs or unpaired
    
    arguments:
        number of pairs of reads (or half the number of total reads)
        folder in which to find fastq
        label for forward reads
        label for reverse reads
    
    output:
        file with paired samples
        file with unpaired samples
            in subsamples folder

    run: parallel --jobs "${PROCESSORS}" "python ${DIR}/libexec/sisrs/sub_sample_for_velvet_unshuff.py ${LEFTREADS} {}" ::: "${FOLDERLISTA[@]}"
"""

import sys
import random
import glob
import os

def main(taxonfolder, left_reads, fileid, fileid2):

    N = int(left_reads)
    #N = int(sys.argv[1])     #number of left reads required per sample for 10x total coverage; (10*genome size) / (read size*2*num_samples)
    samplep,sampleu = [],[]
    i=0

    #taxonfolder=sys.argv[2]
    mainfolder=os.path.dirname(os.path.abspath(taxonfolder))
    taxon_files = glob.glob(taxonfolder+'/*fastq')
    taxon=os.path.basename(taxonfolder)

    #fileid = sys.argv[3]
    #fileid2 = sys.argv[4]

    paired1 = [f for f in taxon_files if fileid in f]
    print('paired files: '+" ".join(paired1))
    unpaired = [f for f in taxon_files if fileid not in f and fileid2 not in f]
    print('unpaired files: '+" ".join(unpaired))

    for fq in paired1:
        f1=open(fq, 'r')
        f2=open(fq.replace(fileid,fileid2), 'r')
        
        for line in f1:
            if not line: break      #have less than desired coverage
            read_info=[line,next(f1),next(f1),next(f1),next(f2),next(f2),next(f2),next(f2)]
            if i < N:        
                samplep.append(read_info)
            elif random.random() < N/float(i+1):
                replace = random.randint(0,N-1)
                samplep[replace] = read_info
            i+=1
        f1.close()
        f2.close()

    for fq in unpaired:
        f1=open(fq, 'r')
        for line in f1:
            if not line: break      #have less than desired coverage
            read_info=[line,next(f1),next(f1),next(f1)]
            try:
                read_info.extend([next(f1),next(f1),next(f1),next(f1)])
            except:
                break
            if i < N:        
                sampleu.append(read_info)
            elif random.random() < N/float(i+1):   
                replace = random.randint(0,N-1)
                if replace<len(samplep):
                    del samplep[replace]
                else:
                    del sampleu[replace-len(samplep)]
                sampleu.append(read_info)
            i+=1
        f1.close()

    if len(samplep)>0:            
        f = open(mainfolder+'/subsamples/'+taxon+'subsampledp.fastq','w')
        for read_info in samplep:
            f.write("".join(read_info))
        f.close()
        print(str(len(samplep))+' paired reads sampled')

    if len(sampleu)>0:
        f = open(mainfolder+'/subsamples/'+taxon+'subsampledu.fastq','w')
        for read_info in sampleu:
            f.write("".join(read_info))
        f.close()
        print(str(len(sampleu)*2)+' unpaired reads sampled')



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.md')) as f:
    long_description = f.read()

setup(
    name='sisrs',
    version='0.1.0',
    packages=find_packages(),
    description="Site Identification from Short Read Sequences.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://cartwrig.ht/software/',
    license='GPLv3',
    install_requires=[
        'biopython',
    ],
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'sisrs-python = sisrs:main'
        ],
    },
    keywords='bioinformatics genetics',
)



import os
from glob import glob
from subprocess import PIPE 
from .aligner import Aligner
from ..process import Process

class Bowtie2Aligner(Aligner):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def index(self, reference_file_path, contig_path_prefix):
        build_command = [
            'bowtie2-build', reference_file_path, contig_path_prefix
        ]
        Process(build_command).wait()

    def align(self, directory, contig_prefix):

        dir_ = directory

        basename = os.path.basename(dir_)

        print("==== Aligning {} as Single-Ended ====".format(basename))
        fastq_filepaths = glob(os.path.join(dir_, '*.fastq'))

        # Generate temp file
        bowtie2_command = [
            'bowtie2',
            '-p', str(self._num_processors),
            '-N', '1',
            '--local',
            '-x', contig_prefix,
            '-U', ','.join(fastq_filepaths)
        ]
        bowtie2_proc = Process(bowtie2_command, stdout=PIPE)

        samtools_to_bam_command = [
            'samtools', 'view', '-Su',
            '-@', str(self._num_processors),
            '-F' , '4',
            '-',
        ]
        samtools_to_bam_proc = Process(
            samtools_to_bam_command, stdin=bowtie2_proc.pipe(),
            stdout=PIPE)

        output_base_path = os.path.join(dir_, basename)
        temp_file_path = output_base_path + '_Temp.bam'

        samtools_sort_command = [
            'samtools', 'sort',
            '-@', str(self._num_processors),
            '-',
            '-o', temp_file_path,
        ]
        samtools_sort_proc = Process(
            samtools_sort_command, stdin=samtools_to_bam_proc.pipe())

        bowtie2_proc.wait()
        samtools_to_bam_proc.wait()
        samtools_sort_proc.wait()


        # Generate header file

        header_file_path = output_base_path + '_Header.sam'

        samtools_header_command = [
            'samtools', 'view',
            '-@', str(self._num_processors),
            '-H', temp_file_path,
            '-o', header_file_path
        ]
        samtools_header_proc = Process(samtools_header_command)

        samtools_header_proc.wait()


        # Generate final output file

        samtools_data_command = [
            'samtools', 'view',
            '-@', str(self._num_processors),
            temp_file_path,
        ]
        samtools_data_proc = Process(samtools_data_command, stdout=PIPE)

        grep_command = [
            'grep', '-v', 'XS:'
        ]
        grep_proc = Process(
            grep_command, stdin=samtools_data_proc.pipe(), stdout=PIPE)

        cat_command = [
            'cat', header_file_path, '-'
        ]
        cat_proc = Process(cat_command, stdin=grep_proc.pipe(), stdout=PIPE)

        samtools_final_command = [
            'samtools', 'view',
            '-@', str(self._num_processors),
            '-b',
            '-',
            '-o', output_base_path + '.bam'
        ]

        samtools_final_proc = Process(
            samtools_final_command, stdin=cat_proc.pipe())

        samtools_data_proc.wait()
        grep_proc.wait()
        cat_proc.wait()
        samtools_final_proc.wait()

        # Remove temp files
        os.remove(temp_file_path)
        os.remove(header_file_path)


